{
  "type": "record",
  "name": "TelemetryClientId",
  "namespace": "com.linkedin.pegasus2avro.telemetry",
  "doc": "A simple wrapper around a String to persist the client ID for telemetry in DataHub's backend DB",
  "fields": [
    {
      "name": "clientId",
      "type": "string",
      "doc": "A string representing the telemetry client ID"
    }
  ],
  "Aspect": {
    "name": "telemetryClientId"
  }
}