{
  "type": "record",
  "name": "DataJobKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a Data Job",
  "fields": [
    {
      "name": "flow",
      "type": "string",
      "doc": "Standardized data processing flow urn representing the flow for the job",
      "Relationship": {
        "entityTypes": [
          "dataFlow"
        ],
        "name": "IsPartOf"
      },
      "Searchable": {
        "fieldName": "dataFlow",
        "fieldType": "URN_PARTIAL",
        "queryByDefault": false
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "jobId",
      "type": "string",
      "doc": "Unique Identifier of the data job",
      "Searchable": {
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "dataJobKey"
  }
}