{
  "type": "record",
  "name": "SourceCode",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "Source Code",
  "fields": [
    {
      "name": "sourceCode",
      "type": {
        "type": "array",
        "items": {
          "type": "record",
          "name": "SourceCodeUrl",
          "doc": "Source Code Url Entity",
          "fields": [
            {
              "name": "type",
              "type": {
                "type": "enum",
                "name": "SourceCodeUrlType",
                "symbols": [
                  "ML_MODEL_SOURCE_CODE",
                  "TRAINING_PIPELINE_SOURCE_CODE",
                  "EVALUATION_PIPELINE_SOURCE_CODE"
                ]
              },
              "doc": "Source Code Url Types"
            },
            {
              "name": "sourceCodeUrl",
              "type": "string",
              "doc": "Source Code Url",
              "java": {
                "class": "com.linkedin.pegasus2avro.common.url.Url",
                "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
              }
            }
          ]
        }
      },
      "doc": "Source Code along with types"
    }
  ],
  "Aspect": {
    "name": "sourceCode"
  }
}